"""Exceptions for this package."""
from typing import List, Any


class FieldNameError(Exception):
    """A field name management error occurred."""


class FieldParentError(Exception):
    """A field-to-field-parent management error occurred."""


class InvalidStructError(Exception):
    """Invalid creation of a custom Struct subclass."""


class StructImplementationError(InvalidStructError):
    """A custom Struct subclass fails to correctly implement required fields."""


class InvalidDataFrameError(Exception):
    """A DataFrame does not match a schema."""


class FieldValueValidationError(Exception):
    """Value is not valid for a field."""


class StructInstantiationError(Exception):
    """Base exception for errors when attempting to instantiate data from a schema."""


class StructInstantiationArgumentsError(StructInstantiationError):
    """Incorrect arguments specified when creating a data object from a schema; e.g., a required argument is missing."""

    def __init__(
        self,
        properties: List[str],
        unfilled_properties: List[str],
        duplicate_properties: List[str],
        surplus_positional_values: List[Any],
        surplus_keyword_args: List[str],
    ):
        msg = ""
        if duplicate_properties:
            msg += (
                "There were struct properties with multiple values. Repeated properties: "
                + ", ".join(duplicate_properties)
                + " \n"
            )
        if surplus_positional_values:
            msg += (
                f"There were {len(surplus_positional_values)} surplus positional arguments. Values for surplus args: "
                + ", ".join(map(str, surplus_positional_values))
                + " \n"
            )
        if surplus_keyword_args:
            msg += "There were surplus keyword arguments: " + ", ".join(surplus_keyword_args) + " \n"
        msg += f"Properties required by this struct are: " + ", ".join(properties)
        if unfilled_properties:
            msg += "\nOmitted struct properties were defaulted to null: " + ", ".join(unfilled_properties)
        super().__init__(msg)


class StructInstantiationArgumentTypeError(StructInstantiationError):
    """
    Found an argument with an incorrect type when creating a data object from a schema.

    For example using a null in a non-nullable or using a string as a float.
    """



"""Struct."""

import dataclasses
from collections import OrderedDict
from dataclasses import dataclass
from difflib import ndiff
from inspect import isclass
from typing import ClassVar, Optional, Mapping, Type, Any, Generator, Tuple, MutableMapping, Dict, List

from pyspark.sql import types as sql_types, DataFrame
from pyspark.sql.types import DataType, StructField, Row

from sparkql.formatters import pretty_schema
from sparkql.schema_builder import schema as schematise
from sparkql.exceptions import (
    InvalidStructError,
    StructImplementationError,
    InvalidDataFrameError,
    StructInstantiationArgumentsError,
    StructInstantiationArgumentTypeError,
    FieldValueValidationError,
)
from sparkql.fields.base import BaseField


@dataclass(frozen=True)
class ValidationResult:
    """
    Describes the result of attempting to validate a struct on a DataFrame's schema.

    Attributes:
        is_valid: Is the DataFrame valid?
        pretty_struct: Human-readable Spark schema for the Struct.
        pretty_data_frame: Human-readable Spark schema for the DataFrame.
        report: A human-readable report on the differences between the two frames. Empty string if no differences.
    """

    is_valid: bool
    pretty_struct: str
    pretty_data_frame: str
    report: str = ""

    def raise_on_invalid(self):
        """
        If the DataFrame did not adhere to the schema, raises an appropriate error.

        Raises:
            InvalidDataFrameError if the DataFrame does not pass validation.
        """
        if not self.is_valid:
            raise InvalidDataFrameError(self.report)


class Struct(BaseField):
    """A struct; shadows StructType in the Spark API."""

    __hash__ = BaseField.__hash__
    _struct_metadata: ClassVar[Optional["_StructInnerMetadata"]] = None

    @classmethod
    def validate_data_frame(cls, dframe: DataFrame) -> ValidationResult:
        """
        Validate that the DataFrame `dframe` adheres to the schema of this struct.

        Args:
            dframe: DataFrame to be validated.

        Returns:
            DataFrameValidationSummary, which describes whether the DataFrame is valid, and additional information
            on the differences.
        """
        schematised_struct = schematise(cls)

        is_valid = dframe.schema == schematised_struct
        pretty_struct = pretty_schema(schematised_struct)
        pretty_dframe = pretty_schema(dframe.schema)
        if is_valid:
            return ValidationResult(is_valid, pretty_struct=pretty_struct, pretty_data_frame=pretty_dframe)

        report = f"Struct schema...\n\n{pretty_struct}\n\n"
        report += f"DataFrame schema...\n\n{pretty_dframe}\n\n"
        report += "Diff of struct -> data frame...\n\n"
        report += "\n".join(
            line_diff
            for line_diff in ndiff(pretty_dframe.splitlines(), pretty_struct.splitlines())
            if not line_diff.startswith("?")
        )
        return ValidationResult(is_valid, pretty_struct=pretty_struct, pretty_data_frame=pretty_dframe, report=report)

    #
    # Handle type management and Spark representations for a Struct object

    @property
    def _spark_type_class(self) -> Type[DataType]:
        return sql_types.StructType

    @property
    def _spark_struct_field(self) -> StructField:
        """The Spark StructField for this field."""
        return StructField(
            name=self._field_name, dataType=self._struct_metadata.spark_struct, nullable=self._is_nullable
        )

    def _validate_on_value(self, value: Any) -> None:
        super()._validate_on_value(value)
        if value is None:
            # super() will have already validate none vs nullability. if None, then it's safe to be none
            return
        if not isinstance(value, Mapping):
            raise FieldValueValidationError(f"Value for a struct must be a mapping, not '{type(value).__name__}'")

        dic: Mapping[str, Any] = value

        fields = list(self._struct_metadata.fields.values())
        field_names = [field._field_name for field in fields]  # pylint: disable=protected-access
        if len(fields) != len(dic):
            raise FieldValueValidationError(
                f"Dict has incorrect number of fields. \n"
                f"Struct requires {len(fields)} fields: {', '.join(field_names)} \n"
                f"Dict has {len(dic)} fields: {', '.join(dic.keys())}"
            )

        if field_names != list(dic.keys()):
            raise FieldValueValidationError(
                f"Dict fields do not match struct fields. \n"
                f"Struct fields: {', '.join(field_names)} \n"
                f"Dict fields: {', '.join(dic.keys())}"
            )

        for field in fields:
            field._validate_on_value(dic[field._field_name])  # pylint: disable=protected-access

    #
    # Hook in to sub-class creation. Ensure fields are pre-processed when a sub-class is declared

    @classmethod
    def __init_subclass__(cls, **options):  # pylint: disable=unused-argument
        """Hook in to the subclassing of this base class; process fields when sub-classing occurs."""
        super().__init_subclass__(**options)  # pytype: disable=attribute-error

        # Ensure a subclass does not break any base class functionality
        for child_prop, child_val in cls.__dict__.items():
            if (child_prop in Struct.__dict__) and (isinstance(child_val, BaseField)):
                raise InvalidStructError(
                    f"Field should not override inherited or reserved class properties: {child_prop}"
                )

        # Extract internal metadata for this class, including parsing the fields
        cls._struct_metadata = _StructInnerMetadata.from_struct_class(cls)

        # Validate the "implements" (if any)
        _Validator(cls).validate()

    #
    # Handle dot chaining for full path ref to nested fields

    def __getattribute__(self, attr_name):
        """
        Customise how field attributes are handled.

        Augment the attribute reference chain to ensure that a field's parent is set.
        """
        attr_value = super().__getattribute__(attr_name)

        if attr_name == "_struct_metadata":
            return attr_value

        resolved_field = self._struct_metadata.resolve_field(
            struct_object=self, attr_name=attr_name, attr_value=attr_value
        )
        if resolved_field is not None:
            return resolved_field

        return attr_value

    #
    # Makers

    @classmethod
    def make_dict(cls, *args, **kwargs) -> Dict[str, Any]:
        """
        Create a data instance of this Struct schema, as a dictionary.

        Specify the values for the fields of this struct using args and keyword args, in the same way you would
        specify values in a constructor. Use the struct's property names are keyword arguments. Omitted fields default
        to null.

        Values are validated according to the struct's schema. Null values are only permitted in nullable fields.

        For example, given a struct:

            ```
            class Article(Struct):
                title = String(name="long_title")
                body = String()
            ```

        then we can write:

            ```
            Article.make_dict(
                title=The title",
                body="The body of the article."
            )
            ```

        to create a dictionary:

            ```
            { "long_title": "The title", "body": "The body of the article."}
            ```

        Raises:
            StructInstantiationArgumentsError:
                If arguments are not specified correctly.
            StructInstantiationArgumentTypeError:
                If a value does not match a field's expected type. Or if attempting to use a null in a non-nullable
                field.

        Returns:
            An instance of this struct, as a dictioniary.
        """
        return _DictMaker(struct_class=cls, positional_args=args, keyword_args=kwargs).make_dict()

    @classmethod
    def make_row(cls, *args, **kwargs) -> Row:
        """Reserved."""
        raise Exception("Function name reserved.")

    #
    # Other methods

    def __eq__(self, other: Any) -> bool:
        """True if `self` equals `other`."""
        return (
            super().__eq__(other)
            and isinstance(other, Struct)
            and self._struct_metadata.fields == other._struct_metadata.fields
            and list(self._struct_metadata.fields.keys()) == list(other._struct_metadata.fields.keys())
        )


@dataclass(frozen=True)
class _StructInnerMetadata:
    """Inner metadata object for Struct classes; hides complexity of field handling."""

    fields: Mapping[str, BaseField]
    # ^ All fields for the Struct, including both native fields and imported from includes

    @staticmethod
    def from_struct_class(struct_class: Type["Struct"]) -> "_StructInnerMetadata":
        """Build instance from a struct class."""
        return _StructInnerMetadata(fields=_FieldsExtractor(struct_class).extract())

    @property
    def spark_struct(self) -> sql_types.StructType:
        """Complete Spark StructType for the sparkql Struct."""
        return sql_types.StructType(
            [field._spark_struct_field for field in self.fields.values()]  # pylint: disable=protected-access
        )

    # Resolving of class attributes
    @staticmethod
    def resolve_field(struct_object: "Struct", attr_name: str, attr_value: Any) -> Optional[BaseField]:
        """
        Attempt to resolve a `getattribute` call on the Struct, returning a BaseField if applicable.

        This should be used to hook into the Struct's `getattribute` behaviour to customise resolution of
        fields. Will also check if the `getattribute` call is attempting to resolve a field. If not, returns None.

        Args:
            struct_object: The instance of the Struct upon which `getattribute` was called.
            attr_name: The name of the `struct_object` attribute to be resolved.
            attr_value: The attribute of the `struct_object`.
        """
        if attr_name.startswith("_"):
            return None

        if isinstance(attr_value, BaseField):
            new_field: BaseField = attr_value._replace_parent(parent=struct_object)  # pylint: disable=protected-access
            return new_field

        return None


#
# Supporting functionality

_META_INNER_CLASS_NAME = "Meta"


@dataclass
class _FieldsExtractor:
    """
    Extracts a Struct's fields from its class, also handling inner Meta, super class, and so on.

    Definitions...

    - Root Struct: The sub-class of Struct being managed by this Handler.
    - Native fields: Fields explicitly declared in the Struct.
    - Meta class: The inner class that can be (optionally) specified within a Struct sub-class. Found at
      `Struct.Meta`.
    - Include Struct: One of (possibly many) Structs specified in the root struct's `Meta.includes`.
    - Includes fields: Fields pulled in from `Meta.includes`.
    """

    INCLUDES_FIELD_NAME: ClassVar[str] = "includes"

    struct_class: Type[Struct]

    def __post_init__(self):
        """Validate input instance variables."""
        _validate_struct_class(self.struct_class)

    def extract(self) -> Mapping[str, BaseField]:
        """Extract the fields."""
        # pylint: disable=attribute-defined-outside-init

        # Extract and hand native fields
        # native field name -> field
        self._native_fields = OrderedDict(self._yield_native_fields())
        for field_name, field in self._native_fields.items():
            field._set_contextual_name(field_name)  # pylint: disable=protected-access

        # Start with super class fields, if any
        self._fields = OrderedDict(self._get_super_class_fields())

        # Add native fields
        self._fields = self._safe_combine_fields(self._fields, self._native_fields)

        # Add includes fields (if any)
        for included_struct in self._yield_included_structs():
            incl_fields = included_struct._struct_metadata.fields  # pylint: disable=protected-access
            for incl_field_name, incl_field in incl_fields.items():
                if incl_field_name not in self._fields:
                    self._fields[incl_field_name] = incl_field  # populate `_fields`
                    # Modify the root struct so that includes fields are available:
                    setattr(self.struct_class, incl_field_name, incl_field)
                elif self._fields[incl_field_name] != incl_field:
                    raise InvalidStructError(
                        "Attempting to replace a field with an 'includes' field of different type. "
                        f"From includes class: {type(included_struct)}. Incompatible field name: {incl_field_name}"
                    )

        return self._fields

    #
    # Extraction and processing of the Struct class

    def _yield_native_fields(self) -> Generator[Tuple[str, Struct], None, None]:
        """
        Get the native fields specified for this Struct class, if any.

        Yields:
            A `(str, BaseField)` pair for each field found in this class. Each pair consists of the attribute name and
            the field.
        """
        for attr_name, attr_value in self.struct_class.__dict__.items():
            if not isinstance(attr_value, BaseField):
                continue
            if attr_name.startswith("_"):
                raise InvalidStructError(
                    f"Fields must not begin with an underscore. Found: {attr_name} = {type(attr_value)}"
                )
            yield (attr_name, attr_value)

    def _yield_included_structs(self) -> Generator[Struct, None, None]:
        """
        Get the Structs specified in the `Meta.includes`, if any.

        A class specified in `Meta.includes` is converted to an instance of the class before returning.

        Yields:
            A Struct object. Note that this is an instance of the Struct, not the class.
            If `Meta` or `Meta.includes` are not provided, no yield.
        """
        for struct in _yield_structs_from_meta(
            self.struct_class, self.INCLUDES_FIELD_NAME  # pytype: disable=wrong-arg-types
        ):
            yield struct

    #
    # Handle inheritance from super class

    def _get_super_class(self) -> Type:
        """Obtain super class."""
        root_class = self.struct_class

        assert len(root_class.__mro__) >= 2
        # by definition, the `_FieldsExtractor` will only be invoked when the `root_class` is subclassing Struct, and
        # thus the MRO for the `root_class` will always have at least two elems

        return root_class.__mro__[1]

    def _get_super_struct_metadata(self) -> Optional[_StructInnerMetadata]:
        """Obtain super class's inner metadata; or None if none found."""
        super_class = self._get_super_class()
        assert super_class is not None and hasattr(super_class, "_struct_metadata")

        if super_class is Struct:
            # we know that the Struct class (i.e., the base for all custom Structs)
            return None

        super_struct_metadata = getattr(super_class, "_struct_metadata")
        assert isinstance(super_struct_metadata, _StructInnerMetadata)
        return super_struct_metadata

    def _get_super_class_fields(self) -> Mapping[str, BaseField]:
        """Returns fields (mapping of attr_name -> field) from the super class; or empty if none found."""
        super_metadata = self._get_super_struct_metadata()
        if super_metadata is None:
            return OrderedDict()
        return OrderedDict(super_metadata.fields)

    #
    # Helpers

    @staticmethod
    def _safe_combine_fields(
        fields_a: Mapping[str, BaseField], fields_b: Mapping[str, BaseField]
    ) -> MutableMapping[str, BaseField]:
        """
        Add fields from `fields_b` into `fields_a`, returning a new copy.

        Only new fields are added. Checks that duplicate fields are identical.
        """
        combined = OrderedDict(fields_a)
        for key, value_b in fields_b.items():
            if key not in combined:
                combined[key] = value_b
            elif combined[key] != value_b:
                raise InvalidStructError(
                    f"Attempting to replace field '{key}' with field of different type: "
                    f"{type(fields_a[key])} vs {type(fields_b[key])}"
                )
        return combined


@dataclass
class _Validator:
    """
    Validate the "implements" requirement of a struct, if any provided.

    Assumes that the inner metadata of the class has been extracted.
    """

    IMPLEMENTS_FIELD_NAME: ClassVar[str] = "implements"

    struct_class: Type[Struct]

    def __post_init__(self):
        """Validate input instance variables."""
        _validate_struct_class(self.struct_class)

    def validate(self):
        """
        Validate that the meets "implements" requirements, if specified.

        Raises:
            StructImplementationError:
                If class does not correctly implement any required structs.
        """
        root_struct_metadata: _StructInnerMetadata = self.struct_class._struct_metadata  # pylint: disable=protected-access
        if root_struct_metadata is None:
            raise ValueError(f"Struct class {self.struct_class} has not had its inner metadata extracted")

        for required_struct in self._yield_implements_structs():

            req_fields = (
                required_struct._struct_metadata.fields  # pylint: disable=protected-access  # pytype: disable=attribute-error
            )
            for req_field_name, req_field in req_fields.items():
                if req_field_name not in root_struct_metadata.fields:  # pytype: disable=attribute-error
                    raise StructImplementationError(
                        f"Struct '{self.struct_class.__name__}' does not implement field '{req_field_name}' "
                        f"required by struct '{type(required_struct).__name__}'"
                    )
                if req_field != root_struct_metadata.fields[req_field_name]:  # pytype: disable=attribute-error
                    # pylint: disable=protected-access
                    raise StructImplementationError(
                        f"Struct '{self.struct_class.__name__}' "  # pytype: disable=attribute-error
                        f"implements field '{req_field_name}' "
                        f"(required by struct '{type(required_struct).__name__}') but field is not compatible. "
                        f"Required {req_field._short_info()} "
                        f"but found {root_struct_metadata.fields[req_field_name]._short_info()}"
                    )

    def _yield_implements_structs(self) -> Generator[Struct, None, None]:
        """Get the Structs specified in the `Meta.implements`, if any."""
        for struct in _yield_structs_from_meta(
            self.struct_class, self.IMPLEMENTS_FIELD_NAME  # pytype: disable=wrong-arg-types
        ):
            yield struct


def _get_inner_meta_class(struct_class: Type[Struct]) -> Optional[Type]:
    """Retrieve the `Meta` inner class of a Struct class, or None if none is provided."""
    if not hasattr(struct_class, _META_INNER_CLASS_NAME):  # pytype: disable=wrong-arg-types
        return None
    inner_meta_class = getattr(struct_class, _META_INNER_CLASS_NAME)

    if not isinstance(inner_meta_class, type):
        raise InvalidStructError(
            f"The '{_META_INNER_CLASS_NAME}' "
            "property of a Struct must only be used as an "
            f"inner class. Found type {type(inner_meta_class)}"
        )
    return inner_meta_class


def _yield_structs_from_meta(source_struct_class: Type[Struct], attribute_name: str) -> Generator[Struct, None, None]:
    """
    Get a list of structs located at an attribute of the Meta inner class, if any.

    A class specified in the list of Structs is converted to an instance of the class before returning.

    Yields:
        A Struct object. Note that this is an instance of the Struct, not the class.
        If `Meta` or the attribute `attribute_name` of `Meta` are not present, no yield.
    """
    inner_meta_class = _get_inner_meta_class(source_struct_class)
    if inner_meta_class is None:
        return

    struct_classes = getattr(inner_meta_class, attribute_name, None)
    if struct_classes is None:
        return

    for index, struct_class in enumerate(struct_classes):
        if not isclass(struct_class):
            raise InvalidStructError(
                f"Encountered non-class item in '{attribute_name}' list of 'Meta' inner class. "
                f"Item at index {index}: '{struct_class}' with type {type(struct_class)}"
            )

        struct_instance = struct_class()
        if not isinstance(struct_instance, Struct):
            raise InvalidStructError(
                "Encountered item in 'includes' list of 'Meta' inner class that is not a Struct or Struct "
                f"subclass. Item at index {index} is {struct_class}"
            )

        yield struct_instance


def _validate_struct_class(struct_class: Type):
    if not issubclass(struct_class, Struct):
        raise ValueError("'struct_class' must inherit from Struct")
    if struct_class is Struct:
        raise ValueError("'struct_class' must not be Struct")


@dataclass
class _DictMaker:
    """Construct an instance of a Struct, as a dictionary."""

    struct_class: Type[Struct]
    positional_args: Tuple[Any]
    keyword_args: Dict[str, Any]

    # internal state
    _struct_property_to_field: Dict[str, BaseField] = dataclasses.field(init=False)

    _property_to_value: "OrderedDict[str, List[Any]]" = dataclasses.field(init=False)
    # ^ store extracted values while we process them. maps property name (i.e., attrib name) to a value.
    #   we'll track if there are any surplus values for the same field

    _surplus_positional_values: List[Any] = dataclasses.field(default_factory=list, init=False)
    _surplus_keyword_args: Dict[str, Any] = dataclasses.field(default_factory=dict, init=False)

    def __post_init__(self):
        # extract `_struct_metadata.fields` for convenience
        inner_meta = self.struct_class._struct_metadata  # pylint: disable=protected-access
        self._struct_property_to_field = inner_meta.fields  # pytype: disable=attribute-error
        self._property_to_value = OrderedDict((property_name, []) for property_name in self._struct_property_to_field)

    def _process_positional_args(self):
        property_to_field = OrderedDict(self._struct_property_to_field)
        args = list(self.positional_args)
        while args:
            arg_value = args.pop(0)
            # determine what field this value is for
            if not property_to_field:
                # we've run out of struct fields. we have surplus positional args
                self._surplus_positional_values.append(arg_value)
            else:
                property_name = next(iter(property_to_field.keys()))
                property_to_field.pop(property_name)
                self._property_to_value[property_name].append(arg_value)

    def _process_keyword_args(self):
        property_to_field = OrderedDict(self._struct_property_to_field)
        kwargs = dict(self.keyword_args)
        while kwargs:
            arg_property_name = next(iter(kwargs.keys()))
            arg_value = kwargs.pop(arg_property_name)

            if arg_property_name not in property_to_field:
                self._surplus_keyword_args[arg_property_name] = arg_value
            else:
                self._property_to_value[arg_property_name].append(arg_value)

    def make_dict(self) -> Dict[str, Any]:
        """Make a dictionary from the given args (see ivars)."""
        # note: "field name" is the concrete field name
        #       "property name" is the class attribute name
        self._process_positional_args()
        self._process_keyword_args()

        #
        # Identify unfilled props and duplicate (repeated) props
        # Note: unfilled_props are not error cases. they will be defaulted to null.
        # duplicate_props *are* error cases
        unfilled_props = []
        duplicate_props = []
        for property_name, values_list in self._property_to_value.items():
            if not values_list:
                unfilled_props.append(property_name)
            elif len(values_list) >= 2:
                duplicate_props.append(property_name)

        #
        # Default unfilled props to null
        for property_name, values_list in self._property_to_value.items():
            if not values_list:
                values_list.append(None)

        #
        # Apply arg handling validation
        if duplicate_props or self._surplus_positional_values or self._surplus_keyword_args:
            raise StructInstantiationArgumentsError(
                properties=list(self._struct_property_to_field.keys()),
                unfilled_properties=unfilled_props,
                duplicate_properties=duplicate_props,
                surplus_positional_values=self._surplus_positional_values,
                surplus_keyword_args=list(self._surplus_keyword_args.keys()),
            )

        #
        # Create instance, with correct ordering, and field name keys
        field_name_to_value = OrderedDict()
        for property_name, field in self._struct_property_to_field.items():
            values = self._property_to_value[property_name]
            assert len(values) == 1, values
            field_name_to_value[field._field_name] = values[0]  # pylint: disable=protected-access

        #
        # Validate arg types
        for field in self._struct_property_to_field.values():
            field_name = field._field_name  # pylint: disable=protected-access
            value = field_name_to_value[field_name]
            try:
                field._validate_on_value(value)  # pylint: disable=protected-access
            except FieldValueValidationError as ex:
                raise StructInstantiationArgumentTypeError(str(ex))

        return dict(field_name_to_value)



"""Accessor functions in order to access field paths, field names, and related field attributes."""

from typing import Sequence

from pyspark.sql import Column
from pyspark.sql.types import StructField

from sparkql.fields.base import BaseField


def path_seq(field: BaseField) -> Sequence[str]:
    """Return the sequence of items that constitute the path to the field `field`; an alias of `BaseField.SEQ`."""
    return field.SEQ


def path_col(field: BaseField) -> Column:
    """Return the Spark column pointing to field `field`; an alias of `BaseField.COL`."""
    return field.COL


def path_str(field: BaseField) -> str:
    """Return dot-delimited path to field `field`; an alias of `BaseField.PATH`."""
    return field.PATH


def name(field: BaseField) -> str:
    """Return field name of field `field`; an alias of `BaseField.NAME`."""
    return field.NAME


def struct_field(field: BaseField) -> StructField:
    """Return the equivalent PySpark StructField of field `field`."""
    return field._spark_struct_field  # pylint: disable=protected-access



"""Concrete atomic fields."""
from datetime import datetime, date
from typing import Type, Any
import decimal

from pyspark.sql.types import (
    ByteType,
    IntegerType,
    LongType,
    ShortType,
    DecimalType,
    DoubleType,
    FloatType,
    StringType,
    BinaryType,
    BooleanType,
    DateType,
    TimestampType,
    DataType,
)

from sparkql.fields.base import AtomicField, IntegralField, FractionalField, _validate_value_type_for_field


#
# DataType -> AtomicType -> NumericType -> IntegralType


class Byte(IntegralField):
    """Field for Spark's ByteType."""

    @property
    def _spark_type_class(self) -> Type[DataType]:
        return ByteType


class Integer(IntegralField):
    """Field for Spark's IntegerType."""

    @property
    def _spark_type_class(self) -> Type[DataType]:
        return IntegerType


class Long(IntegralField):
    """Field for Spark's LongType."""

    @property
    def _spark_type_class(self) -> Type[DataType]:
        return LongType


class Short(IntegralField):
    """Field for Spark's ShortType."""

    @property
    def _spark_type_class(self) -> Type[DataType]:
        return ShortType


#
# DataType -> AtomicType -> NumericType -> FractionalType


class Decimal(FractionalField):
    """Field for Spark's DecimalType."""

    @property
    def _spark_type_class(self) -> Type[DataType]:
        return DecimalType

    def _validate_on_value(self, value: Any) -> None:
        super()._validate_on_value(value)
        _validate_value_type_for_field((decimal.Decimal,), value)


class Double(FractionalField):
    """Field for Spark's DoubleType."""

    @property
    def _spark_type_class(self) -> Type[DataType]:
        return DoubleType

    def _validate_on_value(self, value: Any) -> None:
        super()._validate_on_value(value)
        _validate_value_type_for_field((float,), value)


class Float(FractionalField):
    """Field for Spark's FloatType."""

    @property
    def _spark_type_class(self) -> Type[DataType]:
        return FloatType

    def _validate_on_value(self, value: Any) -> None:
        super()._validate_on_value(value)
        _validate_value_type_for_field((float,), value)


#
# DataType -> AtomicType -> non-numeric types


class String(AtomicField):
    """Field for Spark's StringType."""

    @property
    def _spark_type_class(self) -> Type[DataType]:
        return StringType

    def _validate_on_value(self, value: Any) -> None:
        super()._validate_on_value(value)
        _validate_value_type_for_field((str,), value)


class Binary(AtomicField):
    """Field for Spark's BinaryType."""

    @property
    def _spark_type_class(self) -> Type[DataType]:
        return BinaryType

    def _validate_on_value(self, value: Any) -> None:
        super()._validate_on_value(value)
        _validate_value_type_for_field((bytearray,), value)


class Boolean(AtomicField):
    """Field for Spark's BooleanType."""

    @property
    def _spark_type_class(self) -> Type[DataType]:
        return BooleanType

    def _validate_on_value(self, value: Any) -> None:
        super()._validate_on_value(value)
        _validate_value_type_for_field((bool,), value)


class Date(AtomicField):
    """Field for Spark's DateType."""

    @property
    def _spark_type_class(self) -> Type[DataType]:
        return DateType

    def _validate_on_value(self, value: Any) -> None:
        super()._validate_on_value(value)
        _validate_value_type_for_field((date, datetime), value)


class Timestamp(AtomicField):
    """Field for Spark's TimestampType."""

    @property
    def _spark_type_class(self) -> Type[DataType]:
        return TimestampType

    def _validate_on_value(self, value: Any) -> None:
        super()._validate_on_value(value)
        _validate_value_type_for_field((datetime,), value)



"""Base field and abstract fields."""

from abc import ABC, abstractmethod
from typing import Optional, Type, Any, Tuple, Sequence
import copy

from pyspark.sql import types as sql_type, Column
from pyspark.sql import functions as sql_funcs
from pyspark.sql.types import StructField, DataType

from sparkql.exceptions import FieldNameError, FieldParentError, FieldValueValidationError


# pytype: disable=invalid-annotation


def _validate_value_type_for_field(accepted_types: Tuple[Type, ...], value: Any):
    """Raise error if `value` is not compatible with types; None values are always permitted."""
    if value is not None and not isinstance(value, accepted_types):
        pretty_types = " ,".join("'" + accepted_type.__name__ + "'" for accepted_type in accepted_types)
        raise FieldValueValidationError(
            f"Value '{value}' has invalid type '{value.__class__.__name__}'. Allowed types are: {pretty_types}"
        )


class BaseField(ABC):
    """Root of the field hierarchy; shadows DataType in the Spark API."""

    # Name management logic:
    # - Explicit name (`__name_explicit`): Set via constructor.
    # - Contextual name (`__name_contextual`): Inferred for the field as it is used in a struct object.
    #   This will always get set, although not immediately. The struct object that will contain this field
    #   is responsible for setting the contextual name.
    # The explicit name, if provided, will override the contextual name.

    __nullable: bool = True
    __name_explicit: Optional[str] = None
    __name_contextual: Optional[str] = None
    _parent_struct: Optional["Struct"] = None  # pytype: disable=name-error

    def __init__(self, nullable: bool = True, name: Optional[str] = None):
        """
        Constructor for a base field.

        Args:
            nullable: Is this field nullable.
            name: Field name. If None, field name will be identified via ivar context resolution.
        """
        self.__nullable = nullable
        self.__name_explicit = name

    #
    # Nullability

    @property
    def _is_nullable(self) -> bool:
        """The nullability status of this field."""
        return self.__nullable

    #
    # Field path chaining

    @property
    def _parent(self) -> Optional["Struct"]:  # pytype: disable=name-error
        return self._parent_struct

    def _replace_parent(self, parent: Optional["Struct"] = None) -> "BaseField":  # pytype: disable=name-error
        """Return a copy of this Field with the parent attribute set."""
        field = copy.copy(self)
        if self._parent_struct is not None:
            raise FieldParentError("Attempted to set parent field that has already been set")
        field._parent_struct = parent  # pylint: disable=protected-access
        return field

    def _replace_explicit_name(self, name: Optional[str] = None) -> "BaseField":
        """
        Return a copy of this field, with the explicit name set.

        Should only be used for internal mechanics of handling name resolution during
        path chaining.
        """
        field: BaseField = copy.copy(self)
        if self.__name_explicit is not None:
            raise FieldNameError("Attempted to set an explicit name that has already been set")
        field.__name_explicit = name  # pylint: disable=protected-access
        return field

    #
    # Field name management

    @property
    def _explicit_name(self) -> Optional[str]:
        return self.__name_explicit

    @property
    def _contextual_name(self) -> Optional[str]:
        return self.__name_contextual

    def _set_contextual_name(self, value: str):
        # Intentionally not using an implicit setter here
        if self.__name_contextual is not None:
            raise FieldNameError(
                "Attempted to override a name that has already been set: "
                f"'{value}' replacing '{self.__name_contextual}'"
            )
        self.__name_contextual = value

    @property
    def _field_name(self) -> str:
        """The name for this field."""
        name = self._resolve_field_name()
        if name is None:
            raise FieldNameError(
                "No field name found among: explicit name = {}, inferred name = {}".format(
                    self.__name_explicit, self.__name_contextual
                )
            )
        return name

    def _resolve_field_name(self, default=None) -> Optional[str]:
        """
        Resolve name for this field, or None if no concrete name set.

        Should only be used by this class and its subclasses.
        """
        if self.__name_explicit is not None:
            return self.__name_explicit
        if self.__name_contextual is not None:
            return self.__name_contextual
        return default

    #
    # Public interface to accessing field names and paths

    @property
    def SEQ(self) -> Sequence[str]:
        """
        The sequence of items that constitute the path to this field.

        The result is context-specific and depends on the path to this field through nested structs (if any).
        """
        fields = [self]
        while fields[0]._parent is not None:  # pylint: disable=protected-access
            fields.insert(0, fields[0]._parent)  # pylint: disable=protected-access

        assert all(
            field._resolve_field_name() is not None for field in fields  # pylint: disable=protected-access
        ), f"Encountered an unset name while traversing path. Path is: {_pretty_path(fields)}"

        return [f._field_name for f in fields]  # pylint: disable=protected-access

    @property
    def COL(self) -> Column:
        """
        The Spark column pointing to this field.

        The result is context-specific and depends on the path to this field through nested structs (if any).
        """
        fields_seq = self.SEQ
        col: Column = sql_funcs.col(fields_seq[0])  # pylint: disable=no-member
        for col_field_name in fields_seq[1:]:
            col = col[col_field_name]
        return col

    @property
    def PATH(self) -> str:
        """
        The dot-delimited path to this field.

        The result is context-specific and depends on the path to this field through nested structs (if any).
        """
        return ".".join(self.SEQ)

    @property
    def NAME(self) -> str:
        """The name of this field."""
        return self._field_name

    #
    # Spark type management

    @property
    @abstractmethod
    def _spark_type_class(self) -> Type[DataType]:
        """The class of the Spark type corresponding to this field."""

    @property
    @abstractmethod
    def _spark_struct_field(self) -> StructField:
        """The Spark StructField for this field."""

    @abstractmethod
    def _validate_on_value(self, value: Any) -> None:
        """
        Raises an error if `value` is not compatible with this field.

        Incompatibility may be due to incorrect nullability or incorrect type.
        """
        # for acceptable type declarations according to pytype, see pytype source code:
        #   types.py:1183
        #   _acceptable_types = {...}
        if not self._is_nullable and value is None:
            msg = "Non-nullable field cannot have None value"
            if self._resolve_field_name() is not None:
                msg += f" (field name = '{self._resolve_field_name()}')"
            raise FieldValueValidationError(msg)

    #
    # Misc.

    @abstractmethod
    def __eq__(self, other: Any) -> bool:
        """True if `self` equals `other`."""
        # Subclasses should call this as part of their equality checks
        return (
            isinstance(other, BaseField)
            and self._is_nullable == other._is_nullable
            and self._resolve_field_name() == other._resolve_field_name()  # may be None == None
            and self._spark_type_class == other._spark_type_class
        )

    def __str__(self):
        """Returns the name of this field."""
        # stringifying a field as its field adds some convenience for cases where we need the field
        # name
        return self._resolve_field_name("")

    def _info(self):
        """String formatted object with a more complete summary of this field, primarily for debugging."""
        return (
            f"<{self.__class__.__name__}\n"
            f"  spark type = {self._spark_type_class.__name__}\n"
            f"  nullable = {self._is_nullable}\n"
            f"  name = {self._resolve_field_name()} <- {[self.__name_explicit, self.__name_contextual]}\n"
            f"  parent = {self._parent}\n"
            ">"
        )

    def _short_info(self):
        """Short info string for use in error messages."""
        nullable = "Nullable " if self._is_nullable else ""
        return f"<{nullable}{self.__class__.__name__}: {self._resolve_field_name()}>"

    def __hash__(self):
        return hash((self._is_nullable, self._resolve_field_name(""), self._spark_type_class))

    def __repr__(self):
        return self._short_info()


class AtomicField(BaseField):
    """
    Atomic field type.

    In the Spark API types hierarchy:

    ```
    DataType
     |- AtomicType
     |- ...
    ```
    """

    __hash__ = BaseField.__hash__

    @property
    @abstractmethod
    def _spark_type_class(self) -> Type[DataType]:
        """The class of the Spark type corresponding to this field."""

    @property
    def _spark_data_type(self) -> sql_type.DataType:
        """Corresponding Spark datatype for this class."""
        return self._spark_type_class()

    @property
    def _spark_struct_field(self) -> StructField:
        """The StructField for this object."""
        return StructField(name=self._field_name, dataType=self._spark_data_type, nullable=self._is_nullable)

    def __eq__(self, other: Any) -> bool:
        """True if `self` equals `other`."""
        return (
            super().__eq__(other) and isinstance(other, AtomicField) and self._spark_data_type == other._spark_data_type
        )

    @abstractmethod
    def _validate_on_value(self, value: Any) -> None:
        super()._validate_on_value(value)


class NumericField(AtomicField):
    """
    Numeric field type.

    In the Spark API types hierarchy:

    ```
    DataType
     |- AtomicType
         |- NumericType
     |- ...
    ```
    """

    @property
    @abstractmethod
    def _spark_type_class(self) -> Type[DataType]:
        """The class of the Spark type corresponding to this field."""

    @abstractmethod
    def _validate_on_value(self, value: Any) -> None:
        super()._validate_on_value(value)


class IntegralField(NumericField):
    """
    Integral field type.

    In the Spark API types hierarchy:

    ```
    DataType
     |- AtomicType
         |- NumericType
             |- IntegralType
     |- ...
    ```
    """

    @property
    @abstractmethod
    def _spark_type_class(self) -> Type[DataType]:
        """The class of the Spark type corresponding to this field."""

    def _validate_on_value(self, value: Any) -> None:
        super()._validate_on_value(value)
        _validate_value_type_for_field((int,), value)


class FractionalField(NumericField):
    """
    Fractional field type.

    In the Spark API types hierarchy:

    ```
    DataType
     |- AtomicType
         |- NumericType
             |- FractionalType
     |- ...
    ```
    """

    @property
    @abstractmethod
    def _spark_type_class(self) -> Type[DataType]:
        """The class of the Spark type corresponding to this field."""

    @abstractmethod
    def _validate_on_value(self, value: Any) -> None:
        super()._validate_on_value(value)


def _pretty_path(path: Sequence[BaseField]):
    """Build pretty string of path, for debug and/or error purposes."""
    # pylint: disable=protected-access
    return "< " + " -> ".join(f"'{field._resolve_field_name()}' ({type(field).__name__})" for field in path) + " >"



"""Support for combining Spark schemas."""

from collections import OrderedDict
from copy import copy
from typing import Dict

from pyspark.sql.types import StructType, StructField, ArrayType, DataType


def merge_schemas(struct_a: StructType, struct_b: StructType) -> StructType:
    """
    Merge two schemas and return the merged schema.

    Nested schemas are merged recursively. Fields shared between the two schemas must be compatible; specifically,
    - fields containing atomic types must contain the same type,
    - fields must have the same nullability,
    - arrays must have the same containsNull.

    Args:
        struct_a: A struct to be merged.
        struct_b: A struct to be merged.

    Returns:
        The merger of `struct_a` with `struct_b`.
    """
    return _SchemaMerger.merge_struct_types(struct_a, struct_b)


class _SchemaMerger:
    """
    Merge two schemas.

    See `merge_schemas` for detailed behaviour description.
    """

    @classmethod
    def append_to_fields(cls, fields: Dict[str, StructField], field: StructField) -> None:
        """
        Add `field` to `fields`; modifies `fields` as a side-effect.

        Args:
            fields: Mapping of field name to field.
            field: Field to be added.
        """
        if field.name not in fields:
            fields[field.name] = field
        else:
            fields[field.name] = cls.merge_fields(fields[field.name], field)

    @classmethod
    def merge_fields(cls, field_a: StructField, field_b: StructField) -> StructField:
        assert field_a.name == field_b.name
        if field_a.nullable != field_b.nullable:
            raise ValueError(
                _validation_error_message(
                    "Fields must have matching nullability constraints. "
                    f"nullable of field A is {field_a.nullable}. "
                    f"nullable of field B is {field_b.nullable}",
                    parent_field_name=field_a.name,
                )
            )

        return StructField(
            name=field_a.name,
            dataType=cls.merge_types(field_a.dataType, field_b.dataType, parent_field_name=field_a.name),
            nullable=field_a.nullable,
        )

    #
    # Merge by type

    @classmethod
    def merge_types(cls, type_a: DataType, type_b: DataType, parent_field_name: str) -> DataType:
        """
        Merge two arbitrary types; delegates to corresponding methods.

        `parent_field_name` is the name of the field to which the types belong.
        It is used to generate an intuitive message if validation fails.
        """
        if type(type_a) is not type(type_b):
            raise ValueError(
                _validation_error_message(
                    "Types must match. "
                    f"Type of A is {type_a.__class__.__name__}. Type of B is {type_b.__class__.__name__}",
                    parent_field_name=parent_field_name,
                )
            )

        if isinstance(type_a, StructType):
            assert isinstance(type_b, StructType)
            return cls.merge_struct_types(type_a, type_b)

        if isinstance(type_a, ArrayType):
            assert isinstance(type_b, ArrayType)
            return cls.merge_array_types(type_a, type_b, parent_field_name=parent_field_name)

        return copy(type_a)

    @classmethod
    def merge_struct_types(cls, struct_type_a: StructType, struct_type_b: StructType) -> StructType:
        assert all(isinstance(obj, StructType) for obj in [struct_type_a, struct_type_b])

        fields: Dict[str, StructField] = OrderedDict()
        for field in struct_type_a.fields:
            cls.append_to_fields(fields, field)
        for field in struct_type_b.fields:
            cls.append_to_fields(fields, field)

        return StructType(list(fields.values()))

    @classmethod
    def merge_array_types(cls, array_type_a: ArrayType, array_type_b: ArrayType, parent_field_name) -> ArrayType:
        assert all(isinstance(obj, ArrayType) for obj in [array_type_a, array_type_b])

        if array_type_a.containsNull != array_type_b.containsNull:
            raise ValueError(
                _validation_error_message(
                    "Arrays must have matching containsNull constraints. "
                    f"containsNull of array A is {array_type_a.containsNull}. "
                    f"containsNull of array B is {array_type_b.containsNull}",
                    parent_field_name=parent_field_name,
                )
            )

        return ArrayType(
            elementType=cls.merge_types(
                array_type_a.elementType, array_type_b.elementType, parent_field_name=parent_field_name
            ),
            containsNull=array_type_a.containsNull,
        )


def _validation_error_message(message: str, parent_field_name: str) -> str:
    return f"Cannot merge due to incompatibility in field '{parent_field_name}': {message}"



"""Python Spark SQL DataFrame schema management for sensible humans."""

from sparkql.fields.struct import ValidationResult
from sparkql.schema_builder import schema
from sparkql.accessors import path_col, path_seq, path_str, name, struct_field
from sparkql.formatters import pretty_schema
from sparkql.fields import (
    Byte,
    Integer,
    Long,
    Short,
    Decimal,
    Double,
    Float,
    String,
    Binary,
    Boolean,
    Date,
    Timestamp,
    Array,
    Struct,
)
from sparkql.schema_merger import merge_schemas
from sparkql import exceptions


__all__ = [
    "schema",
    "path_col",
    "path_seq",
    "path_str",
    "name",
    "struct_field",
    "pretty_schema",
    "Byte",
    "Integer",
    "Long",
    "Short",
    "Decimal",
    "Double",
    "Float",
    "String",
    "Binary",
    "Boolean",
    "Date",
    "Timestamp",
    "Array",
    "Struct",
    "ValidationResult",
    "merge_schemas",
    "exceptions",
]



"""Array field."""

import copy
from collections import Sequence
from typing import Optional, Generic, TypeVar, Any

from pyspark.sql.types import ArrayType, StructField

from sparkql.exceptions import FieldValueValidationError
from sparkql.fields.base import BaseField


ArrayElementType = TypeVar("ArrayElementType", bound=BaseField)


class Array(Generic[ArrayElementType], BaseField):
    """
    Array field; shadows ArrayType in the Spark API.

    A spark schema generated with this array field will behave as follows:
    - Nullability of the `StructField` is given by `Array.nullable`, as per normal.
    - `containsNull` of the `ArrayType` is given by the nullability of the element contained
      within this field. In other words, `ArrayType.containsNull` is given by
      `Array.element.nullable`.

    Attributes:
        etype: Data type info for the element of this array. Should be an instance of a `BaseField`.
    """

    __hash__ = BaseField.__hash__

    e: ArrayElementType  # pytype: disable=not-supported-yet  # pylint: disable=invalid-name

    def __init__(self, element: ArrayElementType, nullable: bool = True, name: Optional[str] = None):
        super().__init__(nullable=nullable, name=name)

        if not isinstance(element, BaseField):
            raise ValueError(f"Array element must be a field. Found type: {type(element).__name__}")

        if element._resolve_field_name() is not None:
            # we require that the array's element should be a "vanilla" field. it should not have been given
            # any name (explicit nor contextual)
            raise ValueError(
                "When using a field as the element field of an array, the field should not have a name. "
                f"The field's name resolved to: {element._resolve_field_name()}"
            )

        # hand down this array's explicit name to its child element
        # this is to ensure correct naming in path chaining (see `self._replace_parent` and `path_seq`)
        element = element._replace_explicit_name(name=self._explicit_name)
        self.e = element  # pylint: disable=invalid-name

    #
    # Field path chaining

    def _replace_parent(
        self, parent: Optional["Struct"] = None  # pytype: disable=invalid-annotation,name-error
    ) -> "Struct":  # pytype: disable=invalid-annotation,name-error
        """Return a copy of this array with the parent attribute set."""
        self_copy = copy.copy(self)
        self_copy._parent_struct = parent  # pylint: disable=protected-access
        self_copy.e = self.e._replace_parent(parent=parent)  # pylint: disable=protected-access
        return self_copy

    #
    # Field name management

    def _set_contextual_name(self, value: str):
        super()._set_contextual_name(value)
        # set child to same name as parent; i.e., propagate contextual name downwards:
        self.e._set_contextual_name(value)  # pylint: disable=protected-access

    #
    # Pass through to the element, for users who don't want to use the `.e` field

    def __getattribute__(self, attr_name: str):
        """Custom get attirubte behaviour."""
        if attr_name.startswith("_"):
            return super().__getattribute__(attr_name)

        try:
            attr_value = super().__getattribute__(attr_name)
        except AttributeError:
            attr_value = None

        if attr_value is not None:
            return attr_value

        return getattr(super().__getattribute__("e"), attr_name)

    #
    # Spark type management

    @property
    def _spark_type_class(self):
        return ArrayType

    @property
    def _spark_struct_field(self) -> StructField:
        """The Spark StructField for this field."""
        # containsNull => is used to indicate if elements in a ArrayType value can have null values
        return StructField(
            name=self._field_name,
            dataType=ArrayType(
                # Note that we do not care about the element's field name here:
                elementType=self.e._spark_struct_field.dataType,  # pylint: disable=protected-access
                containsNull=self.e._is_nullable,  # pylint: disable=protected-access
            ),
            nullable=self._is_nullable,
        )

    def _validate_on_value(self, value: Any) -> None:
        super()._validate_on_value(value)
        if value is None:
            # super() will have already validate none vs nullability. if None, then it's safe to be none
            return
        if not isinstance(value, Sequence):
            raise FieldValueValidationError(f"Value for an array must be a sequence, not '{type(value).__name__}'")
        if isinstance(value, str):
            raise FieldValueValidationError(
                f"Value for an array must not be a string. Found value '{value}'. Did you mean to use a list of "
                "strings?"
            )
        for item in value:
            element_field = self.e
            if not element_field._is_nullable and item is None:  # pylint: disable=protected-access
                # to improve readability for errors, we preemptively validate the non-nullability of the array
                # element here
                msg = (
                    "Encountered None value in array, but the element field of this array is specified as "
                    "non-nullable"
                )
                if self._resolve_field_name() is not None:
                    msg += f" (array field name = '{self._resolve_field_name()}')"
                raise FieldValueValidationError(msg)
            self.e._validate_on_value(item)  # pylint: disable=protected-access

    #
    # Misc

    def __eq__(self, other: Any) -> bool:
        """True if `self` equals `other`."""
        return super().__eq__(other) and isinstance(other, Array) and self.e == other.e



"""Schema fields."""

from sparkql.fields.atomic import (
    Byte,
    Integer,
    Long,
    Short,
    Decimal,
    Double,
    Float,
    String,
    Binary,
    Boolean,
    Date,
    Timestamp,
)
from sparkql.fields.array import Array
from sparkql.fields.struct import Struct


__all__ = [
    "Byte",
    "Integer",
    "Long",
    "Short",
    "Decimal",
    "Double",
    "Float",
    "String",
    "Binary",
    "Boolean",
    "Date",
    "Timestamp",
    "Array",
    "Struct",
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sparkql', 'sparkql.fields']

package_data = \
{'': ['*']}

install_requires = \
['pyspark>=2.4.1,<3.0.0']

extras_require = \
{':python_version >= "3.6" and python_version < "3.7"': ['dataclasses>=0.7.0,<0.8.0']}

entry_points = \
{'console_scripts': ['find-releasable-changes = tasks:find_releasable_changes',
                     'lint = tasks:lint',
                     'prepare-release = tasks:prepare_release',
                     'reformat = tasks:reformat',
                     'test = tasks:test',
                     'typecheck = tasks:typecheck',
                     'verify-all = tasks:verify_all']}

setup_kwargs = {
    'name': 'sparkql',
    'version': '0.5.0',
    'description': 'sparkql: Apache Spark SQL DataFrame schema management for sensible humans',
    'long_description': '# sparkql ✨\n\n[![PyPI version](https://badge.fury.io/py/sparkql.svg)](https://badge.fury.io/py/sparkql)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n[![CI](https://github.com/mattjw/sparkql/workflows/CI/badge.svg)](https://github.com/mattjw/sparkql/actions)\n[![codecov](https://codecov.io/gh/mattjw/sparkql/branch/master/graph/badge.svg)](https://codecov.io/gh/mattjw/sparkql)\n\nPython Spark SQL DataFrame schema management for sensible humans.\n\n> _Don\'t sweat it... sparkql it ✨_\n\n## Why use sparkql\n\n`sparkql` takes the pain out of working with DataFrame schemas in PySpark.\nIt makes schema definition more Pythonic. And it\'s\nparticularly useful you\'re dealing with structured data.\n\nIn plain old PySpark, you might find that you write schemas\n[like this](https://github.com/mattjw/sparkql/tree/master/examples/conferences_comparison/plain_schema.py):\n\n```python\nCITY_SCHEMA = StructType()\nCITY_NAME_FIELD = "name"\nCITY_SCHEMA.add(StructField(CITY_NAME_FIELD, StringType(), False))\nCITY_LAT_FIELD = "latitude"\nCITY_SCHEMA.add(StructField(CITY_LAT_FIELD, FloatType()))\nCITY_LONG_FIELD = "longitude"\nCITY_SCHEMA.add(StructField(CITY_LONG_FIELD, FloatType()))\n\nCONFERENCE_SCHEMA = StructType()\nCONF_NAME_FIELD = "name"\nCONFERENCE_SCHEMA.add(StructField(CONF_NAME_FIELD, StringType(), False))\nCONF_CITY_FIELD = "city"\nCONFERENCE_SCHEMA.add(StructField(CONF_CITY_FIELD, CITY_SCHEMA))\n```\n\nAnd then plain old PySpark makes you deal with nested fields like this:\n\n```python\ndframe.withColumn("city_name", df[CONF_CITY_FIELD][CITY_NAME_FIELD])\n```\n\nInstead, with `sparkql`, schemas become a lot\n[more literate](https://github.com/mattjw/sparkql/tree/master/examples/conferences_comparison/sparkql_schema.py):\n\n```python\nclass City(Struct):\n    name = String(nullable=False)\n    latitude = Float()\n    longitude = Float()\n\nclass Conference(Struct):\n    name = String(nullable=False)\n    city = City()\n```\n\nAs does dealing with nested fields:\n\n```python\ndframe.withColumn("city_name", Conference.city.name.COL)\n```\n\nHere\'s a summary of `sparkql`\'s features.\n\n- ORM-like class-based Spark schema definitions.\n- Automated field naming: The attribute name of a field as it appears\n  in its `Struct` is (by default) used as its field name. This name can\n  be optionally overridden.\n- Programatically reference nested fields in your structs with the\n  `PATH` and `COL` special properties. Avoid hand-constructing strings\n  (or `Column`s) to reference your nested fields.\n- Validate that a DataFrame matches a `sparkql` schema.\n- Reuse and build composite schemas with `inheritance`, `includes`, and\n  `implements`.\n- Get a human-readable Spark schema representation with `pretty_schema`.\n- Create an instance of a schema as a dictionary, with validation of\n  the input values.\n\nRead on for documentation on these features.\n\n## Defining a schema\n\nEach Spark atomic type has a counterpart `sparkql` field:\n\n| PySpark type | `sparkql` field |\n|---|---|\n| `ByteType` | `Byte` |\n| `IntegerType` | `Integer` |\n| `LongType` | `Long` |\n| `ShortType` | `Short` |\n| `DecimalType` | `Decimal` |\n| `DoubleType` | `Double` |\n| `FloatType` | `Float` |\n| `StringType` | `String` |\n| `BinaryType` | `Binary` |\n| `BooleanType` | `Boolean` |\n| `DateType` | `Date` |\n| `TimestampType` | `Timestamp` |\n\n`Array` (counterpart to `ArrayType` in PySpark) allows the definition\nof arrays of objects. By creating a subclass of `Struct`, we can\ndefine a custom class that will be converted to a `StructType`.\n\nFor\n[example](https://github.com/mattjw/sparkql/tree/master/examples/arrays/arrays.py),\ngiven the `sparkql` schema definition:\n\n```python\nfrom sparkql import Struct, String, Array\n\nclass Article(Struct):\n    title = String(nullable=False)\n    tags = Array(String(), nullable=False)\n    comments = Array(String(nullable=False))\n```\n\nThen we can build the equivalent PySpark schema (a `StructType`)\nwith:\n\n```python\nfrom sparkql import schema\n\npyspark_struct = schema(Article)\n```\n\nPretty printing the schema with the expression\n`sparkql.pretty_schema(pyspark_struct)` will give the following:\n\n```text\nStructType(List(\n    StructField(title,StringType,false),\n    StructField(tags,\n        ArrayType(StringType,true),\n        false),\n    StructField(comments,\n        ArrayType(StringType,false),\n        true)))\n```\n\n## Features\n\nMany examples of how to use `sparkql` can be found in\n[`examples`](https://github.com/mattjw/sparkql/tree/master/examples).\n\n### Automated field naming\n\nBy default, field names are inferred from the attribute name in the\nstruct they are declared.\n\nFor example, given the struct\n\n```python\nclass Geolocation(Struct):\n    latitude = Float()\n    longitude = Float()\n```\n\nthe concrete name of the `Geolocation.latitude` field is `latitude`.\n\nNames also be overridden by explicitly specifying the field name as an\nargument to the field\n\n```python\nclass Geolocation(Struct):\n    latitude = Float(name="lat")\n    longitude = Float(name="lon")\n```\n\nwhich would mean the concrete name of the `Geolocation.latitude` field\nis `lat`.\n\n### Field paths and nested objects\n\nReferencing fields in nested data can be a chore. `sparkql` simplifies this\nwith path referencing.\n\n[For example](https://github.com/mattjw/sparkql/tree/master/examples/nested_objects/sparkql_example.py), if we have a\nschema with nested objects:\n\n```python\nclass Address(Struct):\n    post_code = String()\n    city = String()\n\n\nclass User(Struct):\n    username = String(nullable=False)\n    address = Address()\n\n\nclass Comment(Struct):\n    message = String()\n    author = User(nullable=False)\n\n\nclass Article(Struct):\n    title = String(nullable=False)\n    author = User(nullable=False)\n    comments = Array(Comment())\n```\n\nWe can use the special `PATH` property to turn a path into a\nSpark-understandable string:\n\n```python\nauthor_city_str = Article.author.address.city.PATH\n"author.address.city"\n```\n\n`COL` is a counterpart to `PATH` that returns a Spark `Column`\nobject for the path, allowing it to be used in all places where Spark\nrequires a column.\n\nFunction equivalents `path_str`, `path_col`, and `name` are also available.\nThis table demonstrates the equivalence of the property styles and the function\nstyles:\n\n| Property style | Function style | Result (both styles are equivalent) |\n| --- | --- | --- |\n| `Article.author.address.city.PATH` | `sparkql.path_str(Article.author.address.city)` | `"author.address.city"` |\n| `Article.author.address.city.COL` | `sparkql.path_col(Article.author.address.city)` | `Column` pointing to `author.address.city` |\n| `Article.author.address.city.NAME` | `sparkql.name(Article.author.address.city)` | `"city"` |\n\nFor paths that include an array, two approaches are provided:\n\n```python\ncomment_usernames_str = Article.comments.e.author.username.PATH\n"comments.author.username"\n\ncomment_usernames_str = Article.comments.author.username.PATH\n"comments.author.username"\n```\n\nBoth give the same result. However, the former (`e`) is more\ntype-oriented. The `e` attribute corresponds to the array\'s element\nfield. Although this looks strange at first, it has the advantage of\nbeing inspectable by IDEs and other tools, allowing goodness such as\nIDE auto-completion, automated refactoring, and identifying errors\nbefore runtime.\n\n### DataFrame validation\n\nStruct method `validate_data_frame` will verify if a given DataFrame\'s\nschema matches the Struct.\n[For example](https://github.com/mattjw/sparkql/tree/master/examples/validation/test_validation.py),\nif we have our `Article`\nstruct and a DataFrame we want to ensure adheres to the `Article`\nschema:\n\n```python\ndframe = spark_session.createDataFrame([{"title": "abc"}])\n\nclass Article(Struct):\n    title = String()\n    body = String()\n```\n\nThen we can can validate with:\n\n```python\nvalidation_result = Article.validate_data_frame(dframe)\n```\n\n`validation_result.is_valid` indicates whether the DataFrame is valid\n(`False` in this case), and `validation_result.report` is a\nhuman-readable string describing the differences:\n\n```text\nStruct schema...\n\nStructType(List(\n    StructField(title,StringType,true),\n    StructField(body,StringType,true)))\n\nDataFrame schema...\n\nStructType(List(\n    StructField(title,StringType,true)))\n\nDiff of struct -> data frame...\n\n  StructType(List(\n-     StructField(title,StringType,true)))\n+     StructField(title,StringType,true),\n+     StructField(body,StringType,true)))\n```\n\nFor convenience,\n\n```python\nArticle.validate_data_frame(dframe).raise_on_invalid()\n```\n\nwill raise a `InvalidDataFrameError` (see `sparkql.exceptions`) if the  \nDataFrame is not valid.\n\n### Creating an instance of a schema\n\n`sparkql` simplifies the process of creating an instance of a struct.\nYou might need to do this, for example, when creating test data, or\nwhen creating an object (a dict or a row) to return from a UDF.\n\nUse `Struct.make_dict(...)` to instantiate a struct as a dictionary.\nThis has the advantage that the input values will be correctly\nvalidated, and it will convert schema property names into their\nunderlying field names.\n\nFor\n[example](https://github.com/mattjw/sparkql/tree/master/examples/struct_instantiation/instantiate_as_dict.py),\ngiven some simple Structs:\n\n```python\nclass User(Struct):\n    id = Integer(name="user_id", nullable=False)\n    username = String()\n\nclass Article(Struct):\n    id = Integer(name="article_id", nullable=False)\n    title = String()\n    author = User()\n    text = String(name="body")\n```\n\nHere are a few examples of creating dicts from `Article`:\n\n```python\nArticle.make_dict(\n    id=1001,\n    title="The article title",\n    author=User.make_dict(\n        id=440,\n        username="user"\n    ),\n    text="Lorem ipsum article text lorem ipsum."\n)\n\n# generates...\n{\n    "article_id": 1001,\n    "author": {\n        "user_id": 440,\n        "username": "user"},\n    "body": "Lorem ipsum article text lorem ipsum.",\n    "title": "The article title"\n}\n```\n\n```python\nArticle.make_dict(\n    id=1002\n)\n\n# generates...\n{\n    "article_id": 1002,\n    "author": None,\n    "body": None,\n    "title": None\n}\n```\n\nSee\n[this example](https://github.com/mattjw/sparkql/tree/master/examples/conferences_extended/conferences.py)\nfor an extended example of using `make_dict`.\n\n### Composite schemas\n\nIt is sometimes useful to be able to re-use the fields of one struct\nin another struct. `sparkql` provides a few features to enable this:\n\n- _inheritance_: A subclass inherits the fields of a base struct class.\n- _includes_: Incorporate fields from another struct.\n- _implements_: Enforce that a struct must implement the fields of\n  another struct.\n\nSee the following examples for a better explanation.\n\n#### Using inheritance\n\nFor [example](https://github.com/mattjw/sparkql/tree/master/examples/composite_schemas/inheritance.py), the following:\n\n```python\nclass BaseEvent(Struct):\n    correlation_id = String(nullable=False)\n    event_time = Timestamp(nullable=False)\n\nclass RegistrationEvent(BaseEvent):\n    user_id = String(nullable=False)\n```\n\nwill produce the following `RegistrationEvent` schema:\n\n```text\nStructType(List(\n    StructField(correlation_id,StringType,false),\n    StructField(event_time,TimestampType,false),\n    StructField(user_id,StringType,false)))\n```\n\n#### Using an `includes` declaration\n\nFor [example](https://github.com/mattjw/sparkql/tree/master/examples/composite_schemas/includes.py), the following:\n\n```python\nclass EventMetadata(Struct):\n    correlation_id = String(nullable=False)\n    event_time = Timestamp(nullable=False)\n\nclass RegistrationEvent(Struct):\n    class Meta:\n        includes = [EventMetadata]\n    user_id = String(nullable=False)\n```\n\nwill produce the `RegistrationEvent` schema:\n\n```text\nStructType(List(\n    StructField(user_id,StringType,false),\n    StructField(correlation_id,StringType,false),\n    StructField(event_time,TimestampType,false)))\n```\n\n#### Using an `implements` declaration\n\n`implements` is similar to `includes`, but does not automatically\nincorporate the fields of specified structs. Instead, it is up to\nthe implementor to ensure that the required fields are declared in\nthe struct.\n\nFailing to implement a field from an `implements` struct will result in\na `StructImplementationError` error.\n\n[For example](https://github.com/mattjw/sparkql/tree/master/examples/composite_schemas/implements.py):\n\n```\nclass LogEntryMetadata(Struct):\n    logged_at = Timestamp(nullable=False)\n\nclass PageViewLogEntry(Struct):\n    class Meta:\n        implements = [LogEntryMetadata]\n    page_id = String(nullable=False)\n\n# the above class declaration will fail with the following StructImplementationError error:\n#   Struct \'RegistrationEvent\' does not implement field \'correlation_id\' required by struct \'EventMetadata\'\n```\n\n\n### Prettified Spark schema strings\n\nSpark\'s stringified schema representation isn\'t very user friendly, particularly for large schemas:\n\n\n```text\nStructType(List(StructField(name,StringType,false),StructField(city,StructType(List(StructField(name,StringType,false),StructField(latitude,FloatType,true),StructField(longitude,FloatType,true))),true)))\n```\n\nThe function `pretty_schema` will return something more useful:\n\n```text\nStructType(List(\n    StructField(name,StringType,false),\n    StructField(city,\n        StructType(List(\n            StructField(name,StringType,false),\n            StructField(latitude,FloatType,true),\n            StructField(longitude,FloatType,true))),\n        true)))\n```\n\n### Merge two Spark `StructType` types\n\nIt can be useful to build a composite schema from two `StructType`s. sparkql provides a\n`merge_schemas` function to do this.\n\n[For example](https://github.com/mattjw/sparkql/tree/master/examples/merge_struct_types/merge_struct_types.py):\n\n```python\nschema_a = StructType([\n    StructField("message", StringType()),\n    StructField("author", ArrayType(\n        StructType([\n            StructField("name", StringType())\n        ])\n    ))\n])\n\nschema_b = StructType([\n    StructField("author", ArrayType(\n        StructType([\n            StructField("address", StringType())\n        ])\n    ))\n])\n\nmerged_schema = merge_schemas(schema_a, schema_b) \n```\n\nresults in a `merged_schema` that looks like:\n\n```text\nStructType(List(\n    StructField(message,StringType,true),\n    StructField(author,\n        ArrayType(StructType(List(\n            StructField(name,StringType,true),\n            StructField(address,StringType,true))),true),\n        true)))\n```\n\n\n## Contributing\n\nContributions are very welcome. Developers who\'d like to contribute to\nthis project should refer to [CONTRIBUTING.md](./CONTRIBUTING.md).\n',
    'author': 'Matt J Williams',
    'author_email': 'mattjw@mattjw.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mattjw/sparkql',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<3.8',
}


setup(**setup_kwargs)



"""Schema management operations."""

from typing import Union, Type

from pyspark.sql.types import StructType


def schema(schema_object: Union["Struct", Type["Struct"]]) -> StructType:  # pytype: disable=name-error
    """Spark schema for a struct object."""
    return schema_object._struct_metadata.spark_struct  # pylint: disable=protected-access
